{******************************************************************************}
{  Pinger class                                                                }
{          M.Kochiashvili                May, 2003                             }
{                                                                              }
{******************************************************************************}
unit MkPinger; interface
uses SysUtils, Classes, Math, Contnrs,
     IdBaseComponent, IdComponent, IdRawBase, IdRawClient, IdIcmpClient;

const
  LogsSubdir = '.\logs\';
  LogsExt    = '.log';

type
  TPingerStat = record
    DateTime  : TDateTime;
    Host      : string[30];
    PkSend    : integer;
    PkReceive : integer;
    PkLossPrc : integer;
    RttMin    :	double;
    RttAvg    : double;
    RttMax    : double;
    Status    : TReplyStatusTypes;
    end;

const
  LogFormat = '%-18s %3d %3d %4d %8.2f %8.2f %8.2f';
type
  TOnFinishPing = procedure(ASender: TComponent; const AStatistics : TPingerStat) of object;

  TMkPinger = class( TIdIcmpClient)
  private
    FPingsToHost: integer;
    FStatistics : TPingerStat;
    FOnFinish : TOnFinishPing;
    FOnEachReply : TOnReplyEvent;
    FFirstPacket : boolean;
    FPinging: boolean;
    procedure SetPingsToHost(const Value: integer);
    procedure ClearStatistics;
    procedure ICMPReply(ASender: TComponent; const ReplyStatus: TReplyStatus);
    procedure SetPinging(const Value: boolean);
  public
    constructor Create(AOwner: TComponent); override;
    property PingsToHost : integer read FPingsToHost write SetPingsToHost default 5;
    procedure PingHost;
    property Statistics : TPingerStat read FStatistics;
    property Pinging : boolean read FPinging write SetPinging;
    property OnFinish : TOnFinishPing read FOnFinish write FOnFinish;
    property OnEachReply : TOnReplyEvent read FOnEachReply write FOnEachReply;
    procedure WriteLog;
    end;

  TMkPingerList = class( TComponentList)
  private
    FActive: boolean;
    FOnFinish: TNotifyEvent;
    FPingsToHost: integer;
    FTimeout: integer;
    function GetItems(Index: Integer): TMkPinger;
    procedure SetItems(Index: Integer; const Value: TMkPinger);
    procedure SetActive(const Value: boolean);
    procedure ItPingFinished(ASender: TComponent; const AStatistics : TPingerStat);
    procedure SetPingsToHost(const Value: integer);
    procedure SetTimeout(const Value: integer);
  public
    constructor Create; overload;
    property Active : boolean read FActive write SetActive;
    function AddNewPing :TMkPinger;
    property PingsToHost : integer read FPingsToHost write SetPingsToHost default 5;
    property Timeout : integer read FTimeout write SetTimeout;
    property PingItems[Index: Integer]: TMkPinger read GetItems write SetItems; default;
    procedure Process;

    property OnFinish : TNotifyEvent read FOnFinish write FOnFinish;
    end;

{ converts Host name to string}
function HostToFileName( H : string) : string;
{}
function FileNameToHost( FN : string) : string;

function ParseLogString( InS : string; var PS : TPingerStat) : boolean;

{ Fill SL with hosts from log subdir}
function GetLogHostsList( SL : TStrings) : integer;


implementation

function HostToFileName( H : string) : string;
var Cou : integer;
begin
  Result := '';
  H := Trim( H);
  if H = '' then exit;
  for Cou := 1 to Length( H) do
    if H[Cou] = '.' then Result := Result + '_'
    else                 Result := Result + H[Cou];
  Result := Result + LogsExt;
end;

function FileNameToHost( FN : string) : string;
var Cou : integer;
begin
  Result := '';
  FN := Trim( FN);
  System.Delete( FN, length( FN) - length( LogsExt) + 1, length( LogsExt));
  for Cou := 1 to length( FN) do begin
    if FN[Cou] = '_' then Result := Result + '.'
    else                  Result := Result + FN[Cou];
    end;
end;

function GetLogHostsList( SL : TStrings) : integer;
var SR : TSearchRec;
begin
  if NOT Assigned( SL) then exit;
  SL.Clear;

  if FindFirst( LogsSubdir + '*' + LogsExt, faAnyFile, SR) = 0 then begin
    repeat
      SL.Add( FileNameToHost( SR.Name));
    until FindNext( SR) <> 0;
    end;
  FindClose( SR);
  Result := SL.Count;
end;

function ParseLogString( InS : string; var PS : TPingerStat) : boolean;
var SPos : integer; S : string;

  procedure SkipBlanks;
  begin
    while ( SPos <= Length( InS)) AND ( InS[ SPos] = ' ') do inc( SPos);
  end;

  function ReadWord : string;
  begin
    Result := '';
    SkipBlanks;
    if SPos > length( InS) then exit;
    while ( SPos <= Length( InS)) AND ( InS[ SPos] <> ' ') do begin
      Result := Result + InS[ SPos];
      inc( SPos);
      end;
  end;

begin
  Result := false;
  FillChar( PS, SizeOf( PS), 0);
  InS := Trim( InS);
  if InS = '' then exit;
  SPos := 1;

  S := ReadWord + ' ' + ReadWord;  PS.DateTime := StrToDateTime( S);
  S := ReadWord;                   PS.PkSend := StrToInt( S);
  S := ReadWord;                   PS.PkReceive := StrToInt( S);
  S := ReadWord;                   PS.PkLossPrc := StrToInt( S);
  S := ReadWord;                   PS.RttMin := StrToFloat( S);
  S := ReadWord;                   PS.RttAvg := StrToFloat( S);
  S := ReadWord;                   PS.RttMax := StrToFloat( S);

  Result := true;
end;

{ TMkPinger }

procedure TMkPinger.ClearStatistics;
begin
  FillChar( FStatistics, SizeOf( TPingerStat), 0);
end;

constructor TMkPinger.Create(AOwner: TComponent);
begin
  inherited;
  FPingsToHost := 5;
  FFirstPacket := true;
  FPinging := false;
  ClearStatistics;
  OnReply := ICMPReply;
end;

procedure TMkPinger.ICMPReply(ASender: TComponent; const ReplyStatus: TReplyStatus);
var RTT : integer;
begin
  if FFirstPacket then begin
    FFirstPacket := false;
    exit;
    end;
  if Assigned( FOnEachReply) then FOnEachReply( ASender, ReplyStatus);
  FStatistics.Host := ReplyStatus.FromIpAddress;
  FStatistics.Status := ReplyStatus.ReplyStatusType;
  case ReplyStatus.ReplyStatusType of
    rsError,
    rsTimeOut,
    rsErrorUnreachable,
    rsErrorTTLExceeded : begin
      with FStatistics do begin
        RttMin := 0;
        RttMax := 0;
        RttAvg := 0;
        end;
      end;
    rsEcho : begin
      with FStatistics do begin
        inc( PkReceive);
        if ReplyStatus.MsRoundTripTime > ReceiveTimeout * 10 then
             RTT := ReceiveTimeout
        else RTT := ReplyStatus.MsRoundTripTime;
        RttMin := Min( RttMin, RTT);
        RttMax := Max( RttMax, RTT);
        RttAvg := RttAvg + RTT;
        end;
      end;
    end;
end;

procedure TMkPinger.PingHost;
var Cou : integer;
begin
  Pinging := true;
  ClearStatistics;
  FFirstPacket := true;
  with FStatistics do begin
    DateTime := Now;
    RttMin := 1.0E15;
    PkSend := FPingsToHost;
    for Cou := 0 to FPingsToHost do begin
      try
        Ping;
      except
        if Assigned( FOnFinish) then FOnFinish( Self, FStatistics);
        Pinging := false;
        end;
      end;
    if PkSend > 0 then PkLossPrc := round(( (PkSend - PkReceive) / PkSend) * 100)
    else               PkLossPrc := 0;
    if PkReceive > 0 then RttAvg := RttAvg / PkReceive
    else                  RttAvg := 0;
    end;
  if Assigned( FOnFinish) then FOnFinish( Self, FStatistics);
  Pinging := false;
end;

procedure TMkPinger.SetPinging(const Value: boolean);
begin
  FPinging := Value;
end;

procedure TMkPinger.SetPingsToHost(const Value: integer);
begin
  FPingsToHost := Value;
end;

procedure TMkPinger.WriteLog;
var S, LFN : string; FS : TFileStream;
begin
  S := Format( LogFormat, [
         DateTimeToStr( Statistics.DateTime),
         Statistics.PkSend,
         Statistics.PkReceive,
         Statistics.PkLossPrc,
         Statistics.RttMin,
         Statistics.RttAvg,
         Statistics.RttMax
             ]);
  LFN := HostToFileName( Host);
  if LFN = '' then exit;
  LFN := LogsSubdir + LFN;
  S := S + #$0D + #$0A;
  if FileExists( LFN) then begin
    FS := TFileStream.Create( LFN, fmOpenReadWrite);
    FS.Seek( 0, soFromEnd);
    end
  else
    FS := TFileStream.Create( LFN, fmCreate);
  try
    FS.Write( S[1], length( S));
  finally
    FS.Free;
    end;
end;

{ TMkPingerList }

function TMkPingerList.AddNewPing: TMkPinger;
begin
  Result := TMkPinger.Create( NIL);
  Result.OnFinish := ItPingFinished;
  Result.ReceiveTimeout := FTimeout;
  Result.PingsToHost := FPingsToHost;
  Add( Result);
end;

constructor TMkPingerList.Create;
begin
  inherited Create( True);
  FPingsToHost := 2;
  FTimeout := 5000;
end;

function TMkPingerList.GetItems(Index: Integer): TMkPinger;
begin
  Result := Items[ Index] as TMkPinger;
end;

procedure TMkPingerList.ItPingFinished(ASender: TComponent; const AStatistics: TPingerStat);
var TotActive : boolean; Cou : integer;
begin
  if Count <= 0 then exit;
  TotActive := false;
  for Cou := 0 to Count - 1 do TotActive := TotActive OR PingItems[ Cou].Pinging;
  if NOT TotActive then begin
    Active := false;
    if Assigned( FOnFinish) then FOnFinish( Self);
    end;
end;

procedure TMkPingerList.Process;
var Cou : integer;
begin
  if Count = 0 then begin
    if Assigned( FOnFinish) then FOnFinish( Self);
    exit;
    end;
  Active := true;
  for Cou := 0 to Count-1 do PingItems[Cou].PingHost;
  if Assigned( FOnFinish) then FOnFinish( Self);
end;

procedure TMkPingerList.SetActive(const Value: boolean);
begin
  if FActive = Value then exit;
  FActive := Value;
//  if FActive then Process;
end;

procedure TMkPingerList.SetItems(Index: Integer; const Value: TMkPinger);
begin
  Items[ Index] := Value;
end;

procedure TMkPingerList.SetPingsToHost(const Value: integer);
var Cou : integer;
begin
  if FPingsToHost = Value then exit;
  FPingsToHost := Value;
  for Cou := 0 to Count-1 do PingItems[Cou].PingsToHost := FPingsToHost;
end;

procedure TMkPingerList.SetTimeout(const Value: integer);
var Cou : integer;
begin
  if FTimeout = Value then exit;
  FTimeout := Value;
  for Cou := 0 to Count-1 do PingItems[Cou].ReceiveTimeout := FTimeout;
end;

initialization
  DateSeparator := '-';
  TimeSeparator := ':';
  ShortDateFormat := 'dd-mm-yy';
  LongTimeFormat := 'hh:nn:ss';
end.
